import java.awt.*;
import java.awt.image.*;
import java.io.*;
import java.util.*;
import javax.imageio.*;
import javax.imageio.plugins.jpeg.*;
import javax.imageio.stream.*;

public class j8phImageCodec {

	/**
	* 画像ファイルのバイナリイメージからピクセル列作成
	*
	* @param data 画像データ
	*
	* @return 作成されたピクセル列。失敗時null;
	*/
	static public int[] decode(byte[] data) {

		if (data == null) {
			return null;
		}

		// 24ビットBMP
		if (data.length >= 28 && data[0] == 'B' && data[1] == 'M' && data[28] == 24) {
			return decodeBMP(data);
		}

		// byte配列を読み込むストリーム作成
		ByteArrayInputStream is = new ByteArrayInputStream(data);

		// デコード結果を格納するBufferedImage宣言
		BufferedImage img = null;

		// デコード
		try {
			img = ImageIO.read(is);
		} catch (Exception e) {
			img = null;
		}

		// デコード失敗ならnullを返す
		if (img == null || img.getWidth() < 1 || img.getHeight() < 1) {
			return null;
		}

		int width = img.getWidth();
		int height = img.getHeight();

		int[] buf = new int[2 + width * height];

		if (buf == null) {
			return null;
		}

		buf[0] = width;
		buf[1] = height;

		for (int i = 0;i < height;i++) {
			for (int j = 0;j < width;j++) {
				buf[2 + j + i * width] = img.getRGB(j, i);
			}
		}

		return buf;

	}

	static public int[] decodeBMP(byte[] bmp) {

		// 24BitBMPでなければ無効
		if (bmp == null || bmp.length < 28 || bmp[0] != 'B' || bmp[1] != 'M' || bmp[28] != 24) {
			return null;
		}

		// ピクセル列までのオフセット
		int offset = (bmp[10] & 0xff) + ((bmp[11] & 0xff) << 8) + ((bmp[12] & 0xff) << 16) + ((bmp[13] & 0xff) << 24);

		if (offset < 10 || offset >= bmp.length) {
			return null;
		}

		int width = (bmp[18] & 0xff) + ((bmp[19] & 0xff) << 8) + ((bmp[20] & 0xff) << 16) + ((bmp[21] & 0xff) << 24);
		int height = (bmp[22] & 0xff) + ((bmp[23] & 0xff) << 8) + ((bmp[24] & 0xff) << 16) + ((bmp[25] & 0xff) << 24);

		if (width < 1 || height < 1) {
			return null;
		}

		/* バッファの１ラインの長さを計算 */
		int length = width * 3;

		if ((width % 4) != 0) {
			length = width * 3 + (4 - (width * 3) % 4);
		}

		if (bmp.length < offset + length * height) {
			return null;
		}

		int[] data = new int[2 + width * height];

		if (data == null) {
			return null;
		}

		data[0] = width;
		data[1] = height;

		// 24bit-32bit変換を行いながらピクセル列設定
		for (int i=0;i<height;i++) {
			for (int j=0;j<width;j++) {

				int r = (int)(bmp[offset + j * 3 + i * length + 2] & 0xff);
				int g = (int)(bmp[offset + j * 3 + i * length + 1] & 0xff);
				int b = (int)(bmp[offset + j * 3 + i * length] & 0xff);

				data[2 + j + (height - i - 1) * width] = (r << 16) & 0x00ff0000 | (g << 8) & 0x0000ff00 | b & 0x000000ff;

			}
		}

		return data;

	}

	static public byte[] encode(int[] data, String format, int quality) {

		if (data == null || data.length < 3 || data[0] < 1 || data[1] < 1) {
			return null;
		}

		if (format == null || format.length() < 1) {
			return null;
		}

		if (format.toLowerCase().equals("bmp")) {
			return encodeBMP(data);
		}

		if (format.toLowerCase().equals("jpg") || format.toLowerCase().equals("jpeg")) {
			return encodeJPG(data, quality);
		}

		BufferedImage img = getImage(data);

		if (img == null) {
			return null;
		}

		ByteArrayOutputStream os = new ByteArrayOutputStream();

		byte[] buf = null;

		try {

			ImageIO.write(img, format, os);

			buf = os.toByteArray();

		} catch (Exception e) {
			buf = null;
		}

		return buf;

	}

	static public byte[] encodeJPG(int[] data, int quality) {

		if (data == null || data.length < 3 || data[0] < 1 || data[1] < 1) {
			return null;
		}

		BufferedImage img = getImage(data);

		if (img == null) {
			return null;
		}

		if (quality < 1) {
			quality = 1;
		}

		if (quality > 100) {
			quality = 100;
		}

		ImageWriter w = null;

		try {
			w = (ImageWriter)ImageIO.getImageWritersByFormatName("jpg").next();
		} catch (Exception e1) {
			w = null;
		}

		if (w == null) {
			return null;
		}

		JPEGImageWriteParam p = new JPEGImageWriteParam(Locale.getDefault());
		p.setCompressionMode(ImageWriteParam.MODE_EXPLICIT);
		p.setCompressionQuality((float)quality / 100.0f);

		ByteArrayOutputStream os = new ByteArrayOutputStream();

		if (os == null) {
			return null;
		}

		ImageOutputStream ios = null;

		try {
			ios = ImageIO.createImageOutputStream(os);
		} catch (Exception e2) {
			ios = null;
		}

		if (ios == null) {
			return null;
		}

		w.setOutput(ios);

		try {
			w.write(null, new IIOImage(img, null ,null), p);
		} catch (Exception e3) {

			try {
				os.close();
			} catch (Exception e){
			}

			return null;

		}

		byte[] buf = os.toByteArray();

		try {
			os.close();
		} catch (Exception e4){
		}

		return buf;

	}

	static public byte[] encodeBMP(int[] img) {

		if (img == null || img.length < 3 || img[0] < 1 || img[1] < 1) {
			return null;
		}

		int width = img[0];
		int height = img[1];

		// ピクセル列の横1ライン幅算出
		int length = width * 3;

		if ((width * 3) % 4 != 0) {
			length = width * 3 + (4 - (width * 3) % 4);
		}

		// データサイズ算出
		int dataSize= 54 + length * height;

		byte data[] = new byte[dataSize];

		if (data == null) {
			return null;
		}

		// バッファにピクセル列作成
		for (int i = 0;i < height;i++) {
			for (int j = 0;j < width;j++) {

				int pixel = img[2 + j + i * width];

				byte r = (byte)(((pixel & 0x00ff0000) >> 16) & 0xff);
				byte g = (byte)(((pixel & 0x0000ff00) >> 8) & 0xff);
				byte b = (byte)(pixel & 0xff);

				data[54 + j * 3 + (height - i - 1) * length] = b;
				data[55 + j * 3 + (height - i - 1) * length] = g;
				data[56 + j * 3 + (height - i - 1) * length] = r;

			}
		}

		// BITMAPFILEHEADER
		data[0]='B';
		data[1]='M';

		// ファイルサイズ記録
		data[2] = (byte)(dataSize & 0xff);
		data[3] = (byte)((dataSize >> 8) & 0xff);
		data[4] = (byte)((dataSize >> 16) & 0xff);
		data[5] = (byte)((dataSize >> 24) & 0xff);

		data[6]=0;
		data[7]=0;
		data[8]=0;
		data[9]=0;

		// ピクセル列までのオフセット記録
		data[10]=54;

		data[11]=0;
		data[12]=0;
		data[13]=0;

		// BITMAPINFOHEADER
		data[14]=40;
		data[15]=0;
		data[16]=0;
		data[17]=0;

		// 画像の横幅記録
		data[18] = (byte)(width & 0xff);
		data[19] = (byte)((width >> 8) & 0xff);
		data[20] = (byte)((width >> 16) & 0xff);
		data[21] = (byte)((width >> 24) & 0xff);

		// 画像の高さ記録
		data[22] = (byte)(height & 0xff);
		data[23] = (byte)((height >> 8) & 0xff);
		data[24] = (byte)((height >> 16) & 0xff);
		data[25] = (byte)((height >> 24) & 0xff);

		data[26]=1;
		data[27]=0;
		data[28]=24;
		data[29]=0;

		return data;

	}

	static int[] getPixels(Image img) {

		if (img == null) {
			return null;
		}

		int width = img.getWidth(null);
		int height = img.getHeight(null);

		if (width < 1 || height < 1) {
			return null;
		}

		BufferedImage tmp = new BufferedImage(width, height, BufferedImage.TYPE_INT_BGR);

		if (tmp == null) {
			return null;
		}

		tmp.getGraphics().drawImage(img, 0, 0, null);

		int[] data = new int[2 + width * height];

		if (data == null) {
			return null;
		}

		data[0] = width;
		data[1] = height;

		for (int i = 0;i < height;i++) {
			for (int j = 0;j < width;j++) {
				data[2 + j + i * width] = tmp.getRGB(j, i);
			}
		}

		return data;

	}

	static BufferedImage getImage(int[] data) {

		if (data == null || data.length < 3 || data[0] < 1 || data[1] < 1) {
			return null;
		}

		int width = data[0];
		int height = data[1];

		if (data.length < 2 + width * height) {
			return null;
		}

		BufferedImage tmp = new BufferedImage(width, height, BufferedImage.TYPE_INT_BGR);

		if (tmp == null) {
			return null;
		}

		for (int i = 0;i < height;i++) {
			for (int j = 0;j < width;j++) {
				tmp.setRGB(j, i, data[2 + j + i * width]);
			}
		}

		return tmp;

	}

}
